from pathlib import Path
from typing import Union
import wszst_yaz0

def guess_bfres_size(file: Union[Path, bytes], name: str = '') -> int:
    real_bytes = file if isinstance(file, bytes) else file.read_bytes()
    if real_bytes[0:4] == b'Yaz0':
        real_bytes = wszst_yaz0.decompress(real_bytes)
    real_size = len(real_bytes)
    del real_bytes
    if name == '':
        if isinstance(file, Path):
            name = file.name
        else:
            raise ValueError('BFRES name must not be blank if passing file as bytes.')
    if '.Tex' in name:
        if real_size < 100:
            return real_size * 9
        elif 100 < real_size <= 2000:
            return real_size * 7
        elif 2000 < real_size <= 3000:
            return real_size * 5
        elif 3000 < real_size <= 4000:
            return real_size * 4
        elif 4000 < real_size <= 8500:
            return real_size * 3
        elif 8500 < real_size <= 12000:
            return real_size * 2
        elif 12000 < real_size <= 17000:
            return int(real_size * 1.75)
        elif 17000 < real_size <= 30000:
            return int(real_size * 1.5)
        elif 30000 < real_size <= 45000:
            return int(real_size * 1.3)
        elif 45000 < real_size <= 100000:
            return int(real_size * 1.2)
        elif 100000 < real_size <= 150000:
            return int(real_size * 1.1)
        elif 150000 < real_size <= 200000:
            return int(real_size * 1.07)
        elif 200000 < real_size <= 250000:
            return int(real_size * 1.045)
        elif 250000 < real_size <= 300000:
            return int(real_size * 1.035)
        elif 300000 < real_size <= 600000:
            return int(real_size * 1.03)
        elif 600000 < real_size <= 1000000:
            return int(real_size * 1.015)
        elif 1000000 < real_size <= 1800000:
            return int(real_size * 1.009)
        elif 1800000 < real_size <= 4500000:
            return int(real_size * 1.005)
        elif 4500000 < real_size <= 6000000:
            return int(real_size * 1.002)
        else:
            return int(real_size * 1.0015)
    else:
        if real_size < 500:
            return real_size * 7
        elif 500 < real_size <= 750:
            return real_size * 4
        elif 750 < real_size <= 2000:
            return real_size * 3
        elif 2000 < real_size <= 400000:
            return int(real_size * 1.75)
        elif 400000 < real_size <= 600000:
            return int(real_size * 1.7)
        elif 600000 < real_size <= 1500000:
            return int(real_size * 1.6)
        elif 1500000 < real_size <= 3000000:
            return int(real_size * 1.5)
        else:
            return int(real_size * 1.25)


def guess_aamp_size(file: Union[Path, bytes], ext: str = '') -> int:
    real_bytes = file if isinstance(file, bytes) else file.read_bytes()
    if real_bytes[0:4] == b'Yaz0':
        real_bytes = wszst_yaz0.decompress(real_bytes)
    real_size = len(real_bytes)
    del real_bytes
    if ext == '':
        if isinstance(file, Path):
            ext = file.suffix
        else:
            raise ValueError('AAMP extension must not be blank if passing file as bytes.')
    ext = ext.replace('.s', '.')
    if ext == '.baiprog':
        if real_size <= 380:
            return real_size * 7
        elif 380 < real_size <= 400:
            return real_size * 6
        elif 400 < real_size <= 450:
            return int(real_size * 5.5)
        elif 450 < real_size <= 600:
            return real_size * 5
        elif 600 < real_size <= 1000:
            return real_size * 4
        elif 1000 < real_size <= 1750:
            return int(real_size * 3.5)
        else:
            return real_size * 3
    elif ext == '.bgparamlist':
        if real_size <= 100:
            return real_size * 20
        elif 100 < real_size <= 150:
            return real_size * 12
        elif 150 < real_size <= 250:
            return real_size * 10
        elif 250 < real_size <= 350:
            return real_size * 8
        elif 350 < real_size <= 450:
            return real_size * 7
        else:
            return real_size * 6
    elif ext == '.bdrop':
        if real_size < 200:
            return int(real_size * 8.5)
        elif 200 < real_size <= 250:
            return real_size * 7
        elif 250 < real_size <= 350:
            return real_size * 6
        elif 350 < real_size <= 450:
            return int(real_size * 5.25)
        elif 450 < real_size <= 850:
            return int(real_size * 4.5)
        else:
            return real_size * 4
    elif ext == '.bxml':
        if real_size < 350:
            return real_size * 6
        elif 350 < real_size <= 450:
            return real_size * 5
        elif 450 < real_size <= 550:
            return int(real_size * 4.5)
        elif 550 < real_size <= 650:
            return real_size * 4
        elif 650 < real_size <= 800:
            return int(real_size * 3.5)
        else:
            return real_size * 3
    elif ext == '.brecipe':
        if real_size < 100:
            return int(real_size * 12.5)
        elif 100 < real_size <= 160:
            return int(real_size * 8.5)
        elif 160 < real_size <= 200:
            return int(real_size * 7.5)
        elif 200 < real_size <= 215:
            return real_size * 7
        else:
            return int(real_size * 6.5)
    elif ext == '.bshop':
        if real_size < 200:
            return int(real_size * 7.25)
        elif 200 < real_size <= 400:
            return real_size * 6
        elif 400 < real_size <= 500:
            return real_size * 5
        else:
            return int(real_size * 4.05)
    elif ext == '.bas':
        if real_size < 100:
            return real_size * 20
        elif 100 < real_size <= 200:
            return int(real_size * 12.5)
        elif 200 < real_size <= 300:
            return real_size * 10
        elif 300 < real_size <= 600:
            return real_size * 8
        elif 600 < real_size <= 1500:
            return real_size * 6
        elif 1500 < real_size <= 2000:
            return int(real_size * 5.5)
        elif 2000 < real_size <= 15000:
            return real_size * 5
        else:
            return int(real_size * 4.5)
    elif ext == '.baslist':
        if real_size < 100:
            return real_size * 15
        elif 100 < real_size <= 200:
            return real_size * 10
        elif 200 < real_size <= 300:
            return real_size * 8
        elif 300 < real_size <= 500:
            return real_size * 6
        elif 500 < real_size <= 800:
            return real_size * 5
        elif 800 < real_size <= 4000:
            return real_size * 4
        else:
            return int(real_size * 3.5)
    elif ext == '.bdmgparam':
        return int(((-0.0018 * real_size) + 6.6273) * real_size) + 500
    else:
        return 0


# pylint: disable=invalid-name,bare-except
""" Functions for building BOTW mods """
from dataclasses import dataclass
from functools import partial
import json
from pathlib import Path
from multiprocessing import Pool, cpu_count
import shutil
import yaml
from zlib import crc32

import aamp
from aamp import yaml_util as ay
import byml
from byml import yaml_util as by
import pymsyt
from rstb import SizeCalculator, ResourceSizeTable
from rstb.util import write_rstb
import sarc
from xxhash import xxh32
from . import AAMP_EXTS, BYML_EXTS, SARC_EXTS, EXEC_DIR, guess, decompress, compress

@dataclass
class BuildParams:
    mod: Path
    out: Path
    content: str
    aoc: str
    be: bool
    guess: bool
    verbose: bool
    hashes: dict

def _is_in_sarc(f: Path) -> bool:
    return any(Path(p).suffix in SARC_EXTS for p in f.parts[:-1])

def _get_canon_name(file: str, allow_no_source: bool = False) -> str:
    name = str(file)\
        .replace("\\", "/")\
        .replace('atmosphere/titles/01007EF00011E000/romfs', 'content')\
        .replace('atmosphere/titles/01007EF00011E001/romfs', 'aoc/0010')\
        .replace('atmosphere/titles/01007EF00011E002/romfs', 'aoc/0010')\
        .replace('atmosphere/titles/01007EF00011F001/romfs', 'aoc/0010')\
        .replace('atmosphere/titles/01007EF00011F002/romfs', 'aoc/0010')\
        .replace('.s', '.')\
        .replace('Content', 'content')\
        .replace('Aoc', 'aoc')
    if 'aoc/' in name:
        return name.replace('aoc/content', 'aoc').replace('aoc', 'Aoc')
    elif 'content/' in name and '/aoc' not in name:
        return name.replace('content/', '')
    elif allow_no_source:
        return name

def _load_rstb(be: bool, file: Path = None) -> ResourceSizeTable:
    table = ResourceSizeTable(b'', be=be)
    if not file:
        ver = 'wiiu' if be else 'switch'
        file = EXEC_DIR / 'data' / ver / 'rstb.json'
    ref_contents = json.loads(file.read_text(), encoding='utf-8')

    def parse_hash(file: str) -> int:
        try:
            return int(file)
        except ValueError:
            return crc32(file.encode('utf8'))

    table.crc32_map = {parse_hash(k): v for k, v in ref_contents['hash_map'].items()}
    table.name_map = {k: v for k, v in ref_contents['name_map'].items()}
    return table

def _get_rstb_val(ext: str, data: bytes, should_guess: bool, be: bool) -> int:
    if not hasattr(_get_rstb_val, 'calc'):
        setattr(_get_rstb_val, 'calc', SizeCalculator())
    val = _get_rstb_val.calc.calculate_file_size_with_ext(data, wiiu=be, ext=ext) # pylint: disable=no-member
    if val == 0 and should_guess:
        if ext in AAMP_EXTS:
            val = guess.guess_aamp_size(data, ext)
        elif ext in {'.bfres', '.sbfres'}:
            val = guess.guess_bfres_size(data, ext)
    return val

def _copy_file(f: Path, params: BuildParams):
    t = params.out / f.relative_to(params.mod)
    if not t.parent.exists():
        t.parent.mkdir(parents=True, exist_ok=True)
    if _is_in_sarc(f):
        shutil.copy(f, t)
        return {}
    else:
        data = f.read_bytes()
        canon = _get_canon_name(f.relative_to(params.mod).as_posix())
        xh = xxh32(
            data if not data[0:4] == b'Yaz0' else decompress(data)
        ).hexdigest()
        t.write_bytes(data)
        if (canon in params.hashes and xh not in params.hashes[canon]) \
           or canon not in params.hashes:
            return {
                canon: _get_rstb_val(t.suffix, data, params.guess, params.be)
            }
        else:
            return {}

def _build_byml(f: Path, be: bool):
    if not hasattr(_build_byml, 'loader'):
        _build_byml.loader = yaml.CLoader
        by.add_constructors(_build_byml.loader)

    with f.open('r', encoding='utf-8') as bf:
        data = yaml.load(bf, Loader=_build_byml.loader)
    file_bytes = byml.Writer(data, be).get_bytes()
    return file_bytes

def _build_aamp(f: Path):
    if not hasattr(_build_aamp, 'loader'):
        _build_aamp.loader = yaml.CLoader
        ay.register_constructors(_build_aamp.loader)

    with f.open('r', encoding='utf-8') as af:
        data = yaml.load(af, Loader=_build_aamp.loader)
    file_bytes = aamp.Writer(data).get_bytes()
    return file_bytes

def _build_yml(f: Path, params: BuildParams):
    rv = {}
    try:
        ext = f.with_suffix('').suffix
        t = params.out / f.relative_to(params.mod).with_suffix('')
        if not t.parent.exists():
            t.parent.mkdir(parents=True)
        data: bytes
        if ext in BYML_EXTS:
            data = _build_byml(f, params.be)
        elif ext in AAMP_EXTS:
            data = _build_aamp(f)
        if not _is_in_sarc(f):
            canon = _get_canon_name(t.relative_to(params.out).as_posix())
            xh = xxh32(data).hexdigest()
            if (canon in params.hashes and xh not in params.hashes[canon]) \
               or canon not in params.hashes:
                return {
                    canon: _get_rstb_val(t.suffix.replace('.s', ''), data, params.guess, params.be)
                }
        t.write_bytes(data if not t.suffix.startswith('.s') else compress(data))
    except Exception as e:
        print(f'Failed to build {f.relative_to(params.mod).as_posix()}: {e}')
        return {}
    else:
        if params.verbose:
            print(f'Built {f.relative_to(params.mod).as_posix()}')
        return rv

def _build_sarc(d: Path, params: BuildParams):
    rvs = {}
    try:
        s = sarc.SARCWriter(params.be)
        lead = ''
        if (d / '.align').exists():
            alignment = int((d / '.align').read_text())
            s.set_default_alignment(alignment)
            (d / '.align').unlink()
        if (d / '.slash').exists():
            lead = '/'
            (d / '.slash').unlink()

        f: Path
        for f in {f for f in d.rglob('**/*') if f.is_file()}:
            path = f.relative_to(d).as_posix()
            data = f.read_bytes()
            xhash = xxh32(data if not data[0:4] == b'Yaz0' \
                          else decompress(data)).hexdigest()
            canon = path.replace('.s', '.')
            if ((canon in params.hashes and xhash not in params.hashes[canon]) \
               or canon not in params.hashes) and not d.suffix in {'.ssarc', '.sarc'}:
                rvs.update({
                    _get_canon_name(path, allow_no_source=True): _get_rstb_val(
                        Path(path).suffix, data, params.guess, params.be
                    )
                })
            s.add_file(lead + path, data)
            f.unlink()

        shutil.rmtree(d)
        sb = s.get_bytes()
        canon = _get_canon_name(d.relative_to(params.out).as_posix())
        if (canon in params.hashes and xxh32(sb).hexdigest() not in params.hashes[canon]) \
           or canon not in params.hashes and not _is_in_sarc(d):
            rvs.update({
                _get_canon_name(d.relative_to(params.out).as_posix()): _get_rstb_val(
                    d.suffix, sb, params.guess, params.be
                )
            })
        d.write_bytes(sb if not (d.suffix.startswith('.s') and d.suffix != '.sarc') \
                      else compress(sb))
    except:
        print(f'Failed to build {d.relative_to(params.out).as_posix()}')
        return {}
    else:
        if params.verbose:
            print(f'Built {d.relative_to(params.out).as_posix()}')
        return rvs

def build_mod(args):
    content = 'content' if args.be else 'atmosphere/titles/01007EF00011E000/romfs'
    aoc = 'aoc' if args.be else 'atmosphere/titles/01007EF00011F001/romfs'
    mod = Path(args.directory)
    if not ((mod / content).exists() or (mod / aoc).exists()):
        print('The specified directory is not valid: no content or aoc folder found')
        exit(1)
    out = mod.with_name(f'{mod.name}_build') if not args.output else Path(args.output)
    if out.exists():
        print('Removing old build...')
        shutil.rmtree(out)

    print('Loading hash table...')
    ver = 'wiiu' if args.be else 'switch'
    hashes = json.loads((EXEC_DIR / 'data' / ver / 'hashes.json').read_text())
    params = BuildParams(mod=mod, out=out, be=args.be, guess=not args.no_guess, 
                         verbose=args.verbose, hashes=hashes, content=content, aoc=aoc)

    print('Scanning source files...')
    files = {f for f in mod.rglob('**/*') if f.is_file()}
    other_files = {f for f in files if f.suffix not in {'.yml', '.msyt'}}
    yml_files = {f for f in files if f.suffix == '.yml'}
    f: Path
    rvs = {}

    print('Copying miscellaneous files...')
    if args.single or len(other_files) < 2:
        for f in other_files:
            rvs.update(_copy_file(f, params))
    else:
        p = Pool(processes=min(len(other_files), cpu_count()))
        results = p.map(partial(_copy_file, params=params), other_files)
        p.close()
        p.join()
        for r in results:
            rvs.update(r)

    if (mod / 'content').exists():
        msg_dirs = {d for d in mod.glob(f'{content}/Pack/Bootup_*.pack') \
                    if d.is_dir() and not d.name == 'Bootup_Graphics.pack'}
        if msg_dirs:
            print('Building MSBT files...')
        for d in msg_dirs:
            msg_dir = next(d.glob('Message/*'))
            new_dir = out / msg_dir.relative_to(mod).with_suffix('.ssarc.ssarc')
            pymsyt.create(msg_dir, new_dir)

    print('Building BYML files...')
    if args.single or len(yml_files) < 2:
        for f in yml_files:
            rvs.update(_build_yml(f, params))
    else:
        p = Pool(processes=min(len(yml_files), cpu_count()))
        results = p.map(partial(_build_yml, params=params), yml_files)
        p.close()
        p.join()
        for r in results:
            rvs.update(r)

    print('Building SARC files...')
    dirs = {d for d in out.rglob('**/*') if d.is_dir()}
    sarc_folders = {d for d in dirs if d.suffix in SARC_EXTS and d.suffix != '.pack'}
    pack_folders = {d for d in dirs if d.suffix == '.pack'}
    if args.single or (len(sarc_folders) + len(pack_folders)) < 3:
        for d in sarc_folders:
            rvs.update(_build_sarc(d, params))
        for d in pack_folders:
            rvs.update(_build_sarc(d, params))
    else:
        sarc_func = partial(_build_sarc, params=params)
        threads = min(len(sarc_folders), cpu_count())
        p = Pool(processes=threads)
        results = p.map(sarc_func, sarc_folders)
        p.close()
        p.join()
        for r in results:
            rvs.update(r)
        p = Pool(processes=threads)
        results = p.map(sarc_func, pack_folders)
        for r in results:
            rvs.update(r)
        p.close()
        p.join()

    if rvs:
        print('Updating RSTB...')
        rp = out / content / 'System' / 'Resource' / 'ResourceSizeTable.product.json'
        table: ResourceSizeTable
        if rp.exists():
            table = _load_rstb(args.be, file=rp)
        else:
            table = _load_rstb(args.be)
        for p, v in rvs.items():
            if not p:
                continue
            msg: str
            if table.is_in_table(p):
                if v > table.get_size(p) > 0:
                    table.set_size(p, v)
                    msg = f'Updated {p} to {v}'
                elif v == 0:
                    table.delete_entry(p)
                    msg = f'Deleted {p}'
                else:
                    msg = f'Skipped {p}'
            else:
                if v > 0 and p not in hashes:
                    table.set_size(p, v)
                    msg = f'Added {p}, set to {v}'
            if args.verbose and msg:
                print(msg)
        write_rstb(table, str(rp.with_suffix('.srsizetable')), args.be)
        rp.unlink()

    print('Mod built successfully')



# pylint: skip-file
import os
from pathlib import Path

SARC_EXTS = {'.sarc', '.pack', '.bactorpack', '.bmodelsh', '.beventpack', '.stera', '.stats',
             '.ssarc', '.spack', '.sbactorpack', '.sbmodelsh', '.sbeventpack', '.sstera', '.sstats'}
AAMP_EXTS = {'.bxml', '.sbxml', '.bas', '.sbas', '.baglblm', '.sbaglblm', '.baglccr', '.sbaglccr',
             '.baglclwd', '.sbaglclwd', '.baglcube', '.sbaglcube', '.bagldof', '.sbagldof',
             '.baglenv', '.sbaglenv', '.baglenvset', '.sbaglenvset', '.baglfila', '.sbaglfila',
             '.bagllmap', '.sbagllmap', '.bagllref', '.sbagllref', '.baglmf', '.sbaglmf',
             '.baglshpp', '.sbaglshpp', '.baiprog', '.sbaiprog', '.baslist', '.sbaslist',
             '.bassetting', '.sbassetting', '.batcl', '.sbatcl', '.batcllist', '.sbatcllist',
             '.bawareness', '.sbawareness', '.bawntable', '.sbawntable', '.bbonectrl',
             '.sbbonectrl', '.bchemical', '.sbchemical', '.bchmres', '.sbchmres', '.bdemo',
             '.sbdemo', '.bdgnenv', '.sbdgnenv', '.bdmgparam', '.sbdmgparam', '.bdrop', '.sbdrop',
             '.bgapkginfo', '.sbgapkginfo', '.bgapkglist', '.sbgapkglist', '.bgenv', '.sbgenv',
             '.bglght', '.sbglght', '.bgmsconf', '.sbgmsconf', '.bgparamlist', '.sbgparamlist',
             '.bgsdw', '.sbgsdw', '.bksky', '.sbksky', '.blifecondition', '.sblifecondition',
             '.blod', '.sblod', '.bmodellist', '.sbmodellist', '.bmscdef', '.sbmscdef', '.bmscinfo',
             '.sbmscinfo', '.bnetfp', '.sbnetfp', '.bphyscharcon', '.sbphyscharcon',
             '.bphyscontact', '.sbphyscontact', '.bphysics', '.sbphysics', '.bphyslayer',
             '.sbphyslayer', '.bphysmaterial', '.sbphysmaterial', '.bphyssb', '.sbphyssb',
             '.bphyssubmat', '.sbphyssubmat', '.bptclconf', '.sbptclconf', '.brecipe', '.sbrecipe',
             '.brgbw', '.sbrgbw', '.brgcon', '.sbrgcon', '.brgconfig', '.sbrgconfig',
             '.brgconfiglist', '.sbrgconfiglist', '.bsfbt', '.sbsfbt', '.bsft', '.sbsft', '.bshop',
             '.sbshop', '.bumii', '.sbumii', '.bvege', '.sbvege', '.bactcapt', '.sbactcapt'}
BYML_EXTS = {'.bgdata', '.sbgdata', '.bquestpack', '.sbquestpack', '.byml', '.sbyml', '.mubin',
             '.smubin', '.baischedule', '.sbaischedule', '.baniminfo', '.sbaniminfo', '.bgsvdata',
             '.sbgsvdata'}
EXEC_DIR = Path(os.path.dirname(os.path.realpath(__file__)))

try:
    import libyaz0.yaz0_cy
    decompress = libyaz0.yaz0_cy.DecompressYaz
    def compress(data: bytes) -> bytes:
        compressed_data = libyaz0.yaz0_cy.CompressYaz(bytes(data), 10)
        result = bytearray(b'Yaz0')
        result += len(data).to_bytes(4, "big")
        result += (0).to_bytes(4, "big")
        result += b'\0\0\0\0'
        result += compressed_data
        return result
except (ImportError, NameError):
    from wszst_yaz0 import decompress, compress



# pylint: disable=invalid-name
""" Functions for unbuilding BOTW mods """
from multiprocessing import Pool, cpu_count
from pathlib import Path
from shutil import rmtree
from typing import Union

import aamp.converters as ac
import byml
from byml import yaml_util as by
import pymsyt
from rstb.util import read_rstb
import sarc
import yaml
from . import AAMP_EXTS, BYML_EXTS, SARC_EXTS, EXEC_DIR, decompress

def _if_unyaz(data: bytes) -> bytes:
    return data if data[0:4] != b'Yaz0' else decompress(data)

def _unbuild_file(f: Path, out: Path, mod: Path, be: bool, verbose: bool):
    of = out / f.relative_to(mod)
    if not of.parent.exists():
        of.parent.mkdir(parents=True, exist_ok=True)
    if f.name == 'ResourceSizeTable.product.srsizetable':
        _unbuild_rstb(f, be, out, mod)
    elif f.suffix in AAMP_EXTS:
        of.with_suffix(f'{f.suffix}.yml').write_bytes(ac.aamp_to_yml(f.read_bytes()))
    elif f.suffix in BYML_EXTS:
        of.with_suffix(f'{f.suffix}.yml').write_bytes(_byml_to_yml(f.read_bytes()))
    elif f.suffix in SARC_EXTS:
        with f.open('rb') as file:
            s = sarc.read_file_and_make_sarc(file)
        if not s:
            return
        _unbuild_sarc(s, of)
        del s
    else:
        of.write_bytes(f.read_bytes())
    if verbose:
        print(f'Unbuilt {f.relative_to(mod).as_posix()}')

def _unbuild_rstb(f: Path, be: bool, out: Path, mod: Path):
    import json
    hash_map = json.loads((EXEC_DIR / 'data' / 'rstb_entries.json').read_text(encoding='utf-8'))
    table = read_rstb(str(f), be)

    def hash_to_name(crc: int) -> Union[str, int]:
        return hash_map[str(crc)] if str(crc) in hash_map else crc

    (out / f.relative_to(mod).with_suffix('.json')).write_text(
        json.dumps({
            'hash_map': {hash_to_name(k): v for k,v in table.crc32_map.items()},
            'name_map': table.name_map
        }, ensure_ascii=False, indent=2)
    )

def _unbuild_sarc(s: sarc.SARC, output: Path):
    SKIP_SARCS = {
        'tera_resource.Cafe_Cafe_GX2.release.ssarc', 'tera_resource.Nin_NX_NVN.release.ssarc'
    }

    output.mkdir(parents=True, exist_ok=True)
    if any(f.startswith('/') for f in s.list_files()):
        (output / '.slash').write_bytes(b'')

    for sf in s.list_files():
        osf = output / sf
        if sf.startswith('/'):
            osf = output / sf[1:]
        osf.parent.mkdir(parents=True, exist_ok=True)
        ext = osf.suffix
        if ext not in {*SARC_EXTS, *AAMP_EXTS, *BYML_EXTS}:
            osf.write_bytes(s.get_file_data(sf).tobytes())
        elif ext in AAMP_EXTS:
            osf.with_suffix(f'{osf.suffix}.yml').write_bytes(
                ac.aamp_to_yml(s.get_file_data(sf).tobytes())
            )
        elif ext in BYML_EXTS:
            osf.with_suffix(f'{osf.suffix}.yml').write_bytes(
                _byml_to_yml(s.get_file_data(sf).tobytes())
            )
        else:
            if osf.name in SKIP_SARCS:
                osf.write_bytes(s.get_file_data(sf).tobytes())
                continue
            try:
                ss = sarc.SARC(_if_unyaz(s.get_file_data(sf).tobytes()))
                _unbuild_sarc(ss, osf)
                del ss
            except ValueError:
                osf.write_bytes(b'')

    if 'Msg_' in output.name:
        pymsyt.export(output, output)
        rmtree(output)
        output.with_suffix('').rename(output)
    if output.suffix in {'.ssarc', '.sarc'}:
        (output / '.align').write_text(str(s.guess_default_alignment()))

def _byml_to_yml(file_bytes: bytes) -> bytes:
    if not hasattr(_byml_to_yml, 'dumper'):
        _byml_to_yml.dumper = yaml.CDumper
        by.add_representers(_byml_to_yml.dumper)
    return yaml.dump(
        byml.Byml(_if_unyaz(file_bytes)).parse(),
        Dumper=_byml_to_yml.dumper
    ).encode('utf8')

def unbuild_mod(args) -> None:
    mod = Path(args.directory)
    if not any(d.exists() for d in {
        mod / 'content', mod / 'aoc', \
        mod / 'atmosphere/titles/01007EF00011E000/romfs',
        mod / 'atmosphere/titles/01007EF00011F001/romfs'
    }):
        print('The specified directory is not valid: no base or DLC folder found')
        exit(1)
    out = mod.with_name(f'{mod.name}_unbuilt') if not args.output else Path(args.output)
    be = (mod / 'content').exists() or (mod / 'aoc').exists()

    print('Analying files...')
    files = {f for f in mod.rglob('**/*') if f.is_file()}
    t = min(len(files), cpu_count())
    print('Unbuilding...')
    if args.single or t < 2:
        for f in files:
            _unbuild_file(f, out, mod, be, args.verbose)
    else:
        from functools import partial
        p = Pool(processes=t)
        p.map(partial(_unbuild_file, mod=mod, be=be, out=out, verbose=args.verbose), files)
        p.close()
        p.join()

    print('Unbuilding complete')



import argparse
import textwrap as _textwrap
import re
from . import unbuilder, builder

def main() -> None:
    """ Main Hyrule Builder function """
    parser = argparse.ArgumentParser(description='Builds and unbuilds BOTW mods for Wii U')
    subparsers = parser.add_subparsers(dest='command', help='Command')
    subparsers.required = True

    b_parser = subparsers.add_parser(
        'build',
        description='Builds a mod into a source-like structure for editing',
        aliases=['b']
    )
    b_parser.add_argument('--be', '-B', help='Use big endian where applicable', action='store_true')
    b_parser.add_argument('--no-guess', '-G', help='Do not use RSTB estimates', action='store_true')
    b_parser.add_argument('--output', '-O', help='Output folder for built mod')
    b_parser.set_defaults(func=builder.build_mod)

    u_parser = subparsers.add_parser(
        'unbuild',
        description='Unbuilds a mod into a source-like structure for editing',
        aliases=['u']
    )
    u_parser.add_argument('--output', '-O', help='Output folder for unbuilt mod')
    u_parser.set_defaults(func=unbuilder.unbuild_mod)

    dir_help = """\
The main mod folder. For Wii U, this must contain a `content` folder and/or an `aoc` folder\
(the latter for DLC files). For Switch, you must use the following layout:
atmosphere
 └─ titles
    ├─ 01007EF00011E000 (for base game files)
    │  └─ romfs
    └─ 01007EF00011F001 (for DLC files)
       └─ romfs"""
    for sp in {b_parser, u_parser}:
        sp.formatter_class = PreserveWhiteSpaceWrapRawTextHelpFormatter
        sp.add_argument('directory', help=dir_help)
        sp.add_argument('--single', '-S', help='Run with single thread', action='store_true')
        sp.add_argument('--verbose', '-V', help='Provide more detailed output', action='store_true')

    args = parser.parse_args()
    args.func(args)

class PreserveWhiteSpaceWrapRawTextHelpFormatter(argparse.RawDescriptionHelpFormatter):
    def __add_whitespace(self, idx, iWSpace, text):
        if idx is 0:
            return text
        return (" " * iWSpace) + text

    def _split_lines(self, text, width):
        textRows = text.splitlines()
        for idx,line in enumerate(textRows):
            search = re.search('\s*[0-9\-]{0,}\.?\s*', line)
            if line.strip() is "":
                textRows[idx] = " "
            elif search:
                lWSpace = search.end()
                lines = [self.__add_whitespace(i,lWSpace,x) for i,x in enumerate(_textwrap.wrap(line, width))]
                textRows[idx] = lines

        return [item for sublist in textRows for item in sublist]

if __name__ == "__main__":
    main()



from setuptools import setup
from pathlib import Path

with open("README.md", "r") as readme:
    long_description = readme.read()

setup(
    name='hyrule_builder',
    version='0.2.1',
    author='NiceneNerd',
    author_email='macadamiadaze@gmail.com',
    description='A mod builder/unbuilder for The Legend of Zelda: Breath of the Wild',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/NiceneNerd/Hyrule-Builder',
    include_package_data=True,
    packages=['hyrule_builder'],
    entry_points={
        'console_scripts': [
            'hyrule_builder = hyrule_builder.__main__:main'
        ]
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 3 :: Only'
    ],
    python_requires='>=3.7',
    install_requires=[
        'aamp>=1.3.0.post1',
        'byml>=2.3.0.post1',
        'pymsyt>=0.1.3',
        'pyyaml>=5.1.2',
        'sarc>=2.0.1',
        'rstb>=1.1.2',
        'wszst-yaz0>=1.2.0.post1',
        'xxhash>=1.3.0'
    ]
)